"""Entrypoint module redirects to command line interface, in case of use of `python -m graphonedrive`.
"""
import sys

from graph_onedrive._cli import main

if __name__ == "__main__":
    sys.exit(main())



"""Graph-OneDrive main functions to create instances of the OneDrive class and save the calss configs.
Funtions: create, create_from_config_file, save_to_config_file.
"""
import json
from pathlib import Path
from typing import Optional
from typing import Union

from graph_onedrive._onedrive import OneDrive


def create(
    client_id: str,
    client_secret: str,
    tenant: str = "common",
    redirect_url: str = "http://localhost:8080",
    refresh_token: Optional[str] = None,
) -> OneDrive:
    """Create an instance of the OneDrive class from arguments.
    Positional arguments:
        client_id (str) -- Azure app client id
        client_secret (str) -- Azure app client secret
    Keyword arguments:
        tenant (str) -- Azure app org tentent id number, use default if multi-tenent (default = "common")
        redirect_url (str)  -- Authentication redirection url (default = "http://localhost:8080")
        refresh_token (str) -- optional token from previous session (default = None)
    Returns:
        onedrive_instance (OneDrive) -- OneDrive object instance
    """

    # Return the OneDrive object instance
    return OneDrive(
        client_id=client_id,
        client_secret=client_secret,
        tenant=tenant,
        redirect_url=redirect_url,
        refresh_token=refresh_token,
    )


def create_from_config_file(
    config_path: Union[str, Path], config_key: str = "onedrive"
) -> OneDrive:
    """Create an instance of the OneDrive class from a config file.
    Positional arguments:
        config_path (str|Path) -- path to configuration json file
    Keyword arguments:
        config_key (str) -- key of the json item storing the configuration (default = "onedrive")
    Returns:
        onedrive_instance (OneDrive) -- OneDrive object instance
    """

    # Read configuration from config file
    print("Reading OneDrive configs...")
    config_path = Path(config_path)
    with open(config_path) as config_file:
        config = json.load(config_file)
    if config_key not in config:
        raise Exception(f"Config Error : Config dict key '{config_key}' incorrect")
    try:
        tenant_id = config[config_key]["tenant_id"]
        client_id = config[config_key]["client_id"]
        client_secret = config[config_key]["client_secret_value"]
    except KeyError:
        raise Exception("Config Error : Config not in acceptable format")
    try:
        redirect_url = config[config_key]["redirect_url"]
    except KeyError:
        redirect_url = "http://localhost:8080"
    try:
        refresh_token = config[config_key]["refresh_token"]
    except KeyError:
        refresh_token = None

    # Create OneDrive object instance
    onedrive_instance = create(
        client_id=client_id,
        client_secret=client_secret,
        tenant=tenant_id,
        redirect_url=redirect_url,
        refresh_token=refresh_token,
    )

    # Get refresh token from instance and update config file
    print("Saving refresh token...")
    with open(config_path) as config_file:
        config = json.load(config_file)
    config[config_key]["refresh_token"] = onedrive_instance.refresh_token
    with open(config_path, "w") as config_file:
        json.dump(config, config_file, indent=4)

    # Return the OneDrive instance
    return onedrive_instance


def save_to_config_file(
    onedrive_instance: OneDrive,
    config_path: Union[str, Path],
    config_key: str = "onedrive",
) -> None:
    """Save the configuration to a json config file.
    Positional arguments:
        onedrive_instance (OneDrive) -- instance with the config to save
        config_path (str|Path) -- path to configuration json file
    Keyword arguments:
        config_key (str) -- key of the json item storing the configuration (default = "onedrive")
    """
    # Read the existing configuration from the file if one exists
    try:
        with open(config_path) as config_file:
            config = json.load(config_file)
    except FileNotFoundError:
        config = {}

    if config_key not in config:
        config[config_key] = {}

    # Set the new configuation
    config[config_key]["tenant_id"] = onedrive_instance._tenant_id
    config[config_key]["client_id"] = onedrive_instance._client_id
    config[config_key]["client_secret_value"] = onedrive_instance._client_secret
    config[config_key]["redirect_url"] = onedrive_instance._redirect
    config[config_key]["refresh_token"] = onedrive_instance.refresh_token

    # Save the configuration to config file
    with open(config_path, "w") as config_file:
        json.dump(config, config_file, indent=4)

    # Nothing returned which signals no errors
    return



"""Command line interface tools for the Python package Graph-OneDrive.
Run terminal command 'graph-onedrive --help' or 'python -m graph-onedrive --help' for details.
"""
import argparse
import json
import os
from typing import Optional
from typing import Sequence
from typing import Tuple

import graph_onedrive._main as graph_onedrive
from graph_onedrive.__init__ import __version__


def main(argv: Optional[Sequence[str]] = None) -> int:
    """Command line interface tools for the Python package Graph OneDrive.
    Use command --help for details.
    """
    # Create the argument parser
    cli_description = "Graph-OneDrive helper functions to create and authenticate configs and interact with OneDrive to test your configuration."
    cli_epilog = "Note: use of the Graph API is subject to the Microsoft terms of use avaliable at https://docs.microsoft.com/en-us/legal/microsoft-apis/terms-of-use"
    parser = argparse.ArgumentParser(description=cli_description, epilog=cli_epilog)
    # Add arguments
    parser_actions = parser.add_argument_group("actions")
    parser_actions.add_argument(
        "-c", "--configure", action="store_true", help="create a new configuration file"
    )
    parser_actions.add_argument(
        "-a",
        "--authenticate",
        action="store_true",
        help="authenticate a configuration file",
    )
    parser_actions.add_argument(
        "-i",
        "--instance",
        action="store_true",
        help="interact with OneDrive to test your config and perform simple tasks",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
        help="Graph-OneDrive version number",
    )
    parser.add_argument(
        "-f",
        "--file",
        action="store",
        type=str,
        metavar="PATH",
        help="optional path to config json file",
    )
    parser.add_argument(
        "-k",
        "--key",
        action="store",
        type=str,
        help="optional config file dictionary key",
    )
    # Parse arguments, using function input args when given for tests
    args = parser.parse_args(argv)

    # Validate arguments
    if not (args.configure or args.authenticate or args.instance):
        parser.error("No action provided, use --help for details")

    if args.file and not args.file.endswith(".json"):
        parser.error("--file path must end in a json file")

    if args.key and args.key == "":
        parser.error("--key provided can not be blank")

    # Call action functions
    if args.configure:
        config(args.file, args.key)
    if args.authenticate:
        authenticate(args.file, args.key)
    if args.instance:
        instance(args.file, args.key)

    # Returning 0 to the terminal to indicate success
    return 0


def config(config_path: Optional[str] = None, config_key: Optional[str] = None) -> None:
    """Create a configuration file."""

    # Set the export directory
    if not config_path:
        if (
            input("Save config.json in current working directory [Y/n]: ")
            .strip()
            .lower()
            == "n"
        ):
            config_path = input("Path to save config (including file name): ").strip()
            if not config_path.endswith(".json"):
                config_path = config_path + ".json"
        else:
            config_path = os.path.join(os.getcwd(), "config.json")

    # Set config dictionary key
    if not config_key:
        if (
            input("Use config dictionary key default 'onedrive' [Y/n]: ")
            .strip()
            .lower()
            == "n"
        ):
            config_key = input("Config dictionary key to use: ").strip()
        else:
            config_key = "onedrive"

    # Load the current file if it exists, otherwsie create dictionary
    if os.path.isfile(config_path):
        with open(config_path) as config_file:
            config = json.load(config_file)
        # For safety do not overwrite existing configs
        if config_key in config:
            raise SystemExit(
                f"Error: '{config_key}' already exists in file, overwrite not permitted."
            )
        config[config_key] = {}
    else:
        config = {config_key: {}}

    # Set basic app credentials
    tenant_id = input("tenant: ").strip()
    client_id = input("client id: ").strip()
    client_secret_value = input("client secret value: ").strip()

    # Set redirect url
    if (
        input("Use redirect url default 'http://localhost:8080' [Y/n]: ")
        .strip()
        .lower()
        == "n"
    ):
        redirect_url = input("Redirect url to use: ").strip()
    else:
        redirect_url = "http://localhost:8080"

    # Format the config into the dictionary
    config[config_key]["tenant_id"] = tenant_id
    config[config_key]["client_id"] = client_id
    config[config_key]["client_secret_value"] = client_secret_value
    config[config_key]["redirect_url"] = redirect_url
    config[config_key]["refresh_token"] = None

    # Save the configuration to config file
    with open(config_path, "w") as config_file:
        json.dump(config, config_file, indent=4)
    print(f"Configuration saved to: {config_path}")


def authenticate(
    config_path: Optional[str] = None, config_key: Optional[str] = None
) -> None:
    """Authenticate with OneDrive and then save the configuration to file."""
    # Get the config file path
    config_path, config_key = get_config_file(config_path, config_key)

    # Create the instance
    onedrive = graph_onedrive.create_from_config_file(
        config_path=config_path, config_key=config_key
    )

    # Save the config
    graph_onedrive.save_to_config_file(
        onedrive_instance=onedrive, config_path=config_path, config_key=config_key
    )
    print(f"Refresh token saved to configuration: {config_path}")


def get_config_file(
    config_path: Optional[str] = None, config_key: Optional[str] = None
) -> Tuple[str, str]:
    """Sets a config path and key by searching the cwd with assistance from from user."""

    if not config_path:
        # Look for json files in the current working directory and confirm with user
        cwd_path = os.getcwd()
        count = 0
        for root, dirs, files in os.walk(cwd_path):
            for file_name in files:
                if file_name.endswith(".json"):
                    if (
                        input(f"Found: {file_name} Use this? [Y/n]: ").strip().lower()
                        != "n"
                    ):
                        config_path = os.path.join(root, file_name)
                        break
                    # Limit number of suggested files to top 5
                    count += 1
                    if count >= 5:
                        break
            else:
                continue
            break
        # Get the file path from the user
        if not config_path:
            while True:
                config_path = input("Path to config file: ").strip()
                if not config_path.endswith(".json"):
                    config_path = config_path + ".json"
                if os.path.isfile(config_path):
                    break
                print("Path could not be validated, please try again.")

    # Open the config file
    with open(config_path) as config_file:
        config = json.load(config_file)

    # Check the config key
    if not config_key:
        config_key = "onedrive"
        if config_key in config:
            if (
                input("Config dictionary key 'onedrive' found. Use this key? [Y/n]: ")
                .strip()
                .lower()
                == "n"
            ):
                config_key = input("Config dictionary key to use: ").strip()
    while config_key not in config:
        print(f"Config dictionary key '{config_key}' not found.")
        config_key = input("Config dictionary key to use: ").strip()

    return config_path, config_key


def instance(
    config_path: Optional[str] = None, config_key: Optional[str] = None
) -> None:
    """Interact directly with OneDrive in the command line to perform simple tasks and test the configuration."""

    # Check with user if using config file
    if config_path or config_key:
        use_config_file = True
    elif input("Load an existing config file [Y/n]: ").strip().lower() != "n":
        use_config_file = True
    else:
        use_config_file = False

    # Create the instance
    if use_config_file:
        # Get the config details
        config_path_verified, config_key_verified = get_config_file(
            config_path, config_key
        )

        # Create session
        onedrive = graph_onedrive.create_from_config_file(
            config_path=config_path_verified, config_key=config_key_verified
        )
    else:
        print("Manual configuration entry:")
        client_id = input("client_id: ").strip()
        client_secret = input("client_secret: ").strip()
        tenant = input("tenant (leave blank to use 'common': ").strip()
        if tenant == "":
            tenant = "common"
        refresh_token = input("refresh_token (leave blank to reauthenticate): ").strip()
        onedrive = graph_onedrive.create(
            client_id, client_secret, tenant, refresh_token
        )

    # Present menu to user and trigger commands
    try:
        help_info = """
    Graph-OneDrive cli instance actions:
    u / usage      :  prints the OneDrive usage
    od / onedrive  :  print the OneDrive details
    li / list      :  list the contents of a folder
    de / detail    :  print metadata of an item
    md / mkdir     :  make a new folder
    mv / move      :  move an item
    cp / copy      :  copy an item
    rn / rename    :  rename an item
    rm / remove    :  remove an item
    dl / download  :  download a file
    ul / upload    :  upload a file
    exit / quit    :  exit the menu\n"""
        print(help_info)
        while True:
            command = input("Please enter command: ").strip().lower()

            if command == "help":
                print(help_info)

            elif command in ["u", "usage"]:
                onedrive.get_usage(verbose=True)

            elif command in ["li", "list"]:
                folder_id: Optional[str] = input(
                    "Folder id to look into (enter nothing for root): "
                ).strip()
                if folder_id == "":
                    folder_id = None
                onedrive.list_directory(folder_id, verbose=True)

            elif command in ["d", "detail"]:
                item_id = input("Item id to detail: ").strip()
                onedrive.detail_item(item_id, verbose=True)

            elif command in ["md", "mkdir"]:
                parent_folder_id: Optional[str] = input(
                    "Parent folder id (enter nothing for root): "
                ).strip()
                if parent_folder_id == "":
                    parent_folder_id = None
                folder_name = input("Name of the new folder: ").strip()
                response = onedrive.make_folder(folder_name, parent_folder_id)
                print(response)

            elif command in ["mv", "move"]:
                item_id = input("Item id of the file/folder to move: ").strip()
                new_folder_id = input("New parent folder id: ").strip()
                if input("Specify a new file name? [y/N]: ").strip().lower() == "y":
                    new_file_name: Optional[str] = input(
                        "New file name (with extension): "
                    ).strip()
                else:
                    new_file_name = None
                response = onedrive.move_item(item_id, new_folder_id, new_file_name)
                print(f"Item {item_id} was moved to {new_folder_id}")

            elif command in ["cp", "copy"]:
                item_id = input("Item id of the file/folder to copy: ").strip()
                new_folder_id = input("New parent folder id: ").strip()
                if input("Specify a new name? [y/N]: ").strip().lower() == "y":
                    new_file_name = input("New file name (with extension): ").strip()
                else:
                    new_file_name = None
                response = onedrive.copy_item(item_id, new_folder_id, new_file_name)
                print(
                    f"Item was copied to folder {new_folder_id} with new item id {response}"
                )

            elif command in ["rn", "rename"]:
                item_id = input("Item id of the file/folder to rename: ").strip()
                new_file_name = input("New file name (with extension): ").strip()
                response = onedrive.rename_item(item_id, new_file_name)
                print("Item was renamed.")

            elif command in ["rm", "remove", "delete"]:
                item_id = input("Item id of the file/folder to remove: ").strip()
                response = onedrive.delete_item(item_id)
                if response == True:
                    print(f"Item {item_id} was successfully removed.")

            elif command in ["dl", "download"]:
                item_id = input("Item id to download: ").strip()
                print("Downloading...")
                response = onedrive.download_file(item_id)
                print(
                    f"Item was downloaded in the current working directory as {response}"
                )

            elif command in ["ul", "upload"]:
                file_path_input = input("Provide full file path: ").strip()
                file_path = file_path_input
                if input("Rename file? [y/N]: ").strip().lower() == "y":
                    new_file_name = input(
                        "Upload as file name (with extension): "
                    ).strip()
                else:
                    new_file_name = None
                parent_folder_id = input(
                    "Folder id to upload within (enter nothing for root): "
                ).strip()
                if parent_folder_id == "":
                    parent_folder_id = None
                response = onedrive.upload_file(
                    file_path, new_file_name, parent_folder_id
                )
                print(response)

            elif command in ["od", "onedrive", "drive"]:
                print("Drive id:    ", onedrive._drive_id)
                print("Drive name:  ", onedrive._drive_name)
                print("Drive type:  ", onedrive._drive_type)
                print("Owner id:    ", onedrive._owner_id)
                print("Owner email: ", onedrive._owner_email)
                print("Owner name:  ", onedrive._owner_name)
                print(
                    "Quota used:  ",
                    round(onedrive._quota_used / (1024 * 1024), 2),
                    "mb",
                )
                print(
                    "Quota remain:",
                    round(onedrive._quota_remaining / (1024 * 1024), 2),
                    "mb",
                )
                print(
                    "Quota total: ",
                    round(onedrive._quota_total / (1024 * 1024), 2),
                    "mb",
                )

            elif command == "_access":
                print(onedrive._access_token)

            elif command == "_refresh":
                print(onedrive.refresh_token)

            elif command in ["exit", "exit()", "quit", "q", "end"]:
                if use_config_file:
                    graph_onedrive.save_to_config_file(
                        onedrive,
                        config_path=config_path_verified,
                        config_key=config_key_verified,
                    )
                break

            else:
                print("Command not recognised. Use 'help' for info or 'exit' to quit.")

    except KeyboardInterrupt:
        if use_config_file:
            graph_onedrive.save_to_config_file(
                onedrive,
                config_path=config_path_verified,
                config_key=config_key_verified,
            )


if __name__ == "__main__":
    raise SystemExit(main())



__version__ = "0.0.1a10"

# 1.2.0.dev1  # Development release
# 1.2.0a1     # Alpha Release
# 1.2.0b1     # Beta Release
# 1.2.0rc1    # Release Candidate
# 1.2.0       # Final Release

from graph_onedrive._main import *



"""Contains the OneDrive object class to interact through the Graph API using the Python package Graph-OneDrive.
"""
import functools
import json
import os
import re
import secrets
import shutil
import urllib.parse
import warnings
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from time import sleep
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import requests

from graph_onedrive._decorators import token_required


class OneDrive:
    """Creates an instance to interact with Microsoft's OneDrive platform through their Graph API.
    Positional arguments:
        client_id (str)     -- Azure app client id
        client_secret (str) -- Azure app client secret
    Keyword arguments:
        tenant (str)        -- Azure app org tenant id number, use default if multi-tenant (default = "common")
        redirect_url (str)  -- Authentication redirection url (default = "http://localhost:8080")
        refresh_token (str) -- optional token from previous session (default = None)
    Attributes:
        refresh_token (str) -- single-use token to supply when recreating the instance to skip authorization
    Methods:
        get_usage           -- account current usage and total capacity
        list_directory      -- lists all of the items and their attributes within a directory
        detail_item         -- get item details
        make_folder         -- creates a folder
        move_item           -- moves an item
        copy_item           -- copies an item
        rename_item         -- renames an item
        delete_item         -- deletes an item
        download_file       -- downloads a file to the working directory
        upload_file         -- uploads a file as a single piece if small and otherwise calls upload_large_file
        upload_large_file   -- uploads a file in chunks, typically calling upload_file is suggested
    """

    # Set class constants for the Graph API
    _API_VERSION = "v1.0"
    _API_URL = "https://graph.microsoft.com/" + _API_VERSION + "/"
    _AUTH_BASE_URL = "https://login.microsoftonline.com/"
    _AUTH_ENDPOINT = "/oauth2/v2.0/"

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        tenant: str = "common",
        redirect_url: str = "http://localhost:8080",
        refresh_token: Optional[str] = None,
    ) -> None:
        # Set private attributes
        self._client_id = client_id
        self._client_secret = client_secret
        self._tenant_id = tenant
        self._auth_url = self._AUTH_BASE_URL + self._tenant_id + self._AUTH_ENDPOINT
        self._scope = "files.readwrite"
        self._redirect = redirect_url
        self._access_token = ""
        self._access_expires = 0.0
        # Set public attributes
        if refresh_token:
            self.refresh_token: str = refresh_token
        else:
            self.refresh_token = ""
        # Initiate generation of authorization tokens
        self._get_token()
        self._create_headers()
        self._get_drive_details()

    def _get_token(self) -> None:
        """INTERNAL: Get access and refresh tokens from the Graph API.
        Calls get_authorization function if an existing refresh token (from a previous session) is not provided.
        """

        # Generate request body
        request_url = self._auth_url + "token"
        body = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "scope": "offline_access " + self._scope,
            "redirect_uri": self._redirect,
        }

        # Set grant type
        # If no refresh token provided, get new authorization code
        if self.refresh_token != "":
            body["grant_type"] = "refresh_token"
            body["refresh_token"] = self.refresh_token
        else:
            authorization_code = self._get_authorization()
            body["grant_type"] = "authorization_code"
            body["code"] = authorization_code

        # Make the request
        response = requests.post(request_url, data=body)
        status_code = response.status_code
        response_data = json.loads(response.text)

        # Check response was okay
        if status_code != 200:
            error_descrpition = response_data.get(
                "error_description", "Unknown error - no error description returned"
            )
            raise Exception(f"API Error : {error_descrpition} ")

        # Set the access and refresh tokens to the instance attributes
        try:
            self._access_token = response_data["access_token"]
        except KeyError:
            raise Exception("Response Error : Response did not return an access token")
        try:
            self.refresh_token = response_data["refresh_token"]
        except KeyError:
            warnings.warn(
                "Response Warn : Response did not return a refresh token, existing config not updated",
                stacklevel=2,
            )

        # Set an expiry time, removing 60 seconds assumed for processing
        expires = response_data["expires_in"] - 60
        expires = datetime.now() + timedelta(seconds=expires)
        self._access_expires = datetime.timestamp(expires)

    def _get_authorization(self) -> str:
        """INTERNAL: Get authorization code by generating a url for the user to authenticate and authorize the app with.
        The user then return the response manually for the function to then extract the authorization code from.
        The authorization code has a short life and should only be used to generate access and refresh tokens.
            Returns:
                authorization_code (str) -- Graph API authorization code valid once for about 10 mins
        """

        # Create state used for check
        alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"  # = string.ascii_letters + string.digits
        state = "".join(secrets.choice(alphabet) for i in range(10))

        # Generate request url
        request_url = self._auth_url + "authorize"
        request_url += "?client_id=" + self._client_id
        request_url += "&response_type=code"
        request_url += "&redirect_uri=" + urllib.parse.quote_plus(self._redirect)
        request_url += "&response_mode=query"
        request_url += "&scope=offline_access%20" + self._scope
        request_url += "&state=" + state

        # Make request (manually)
        print("Manual app authorization required.")
        print("Step 1: Copy the below URL and paste into a web browser.")
        print("AUTHORIZATION URL --------------")
        print(request_url)
        print("--------------------------------")
        print("Step 2: Authorize the app using your account.")
        print("You will be redirected (potentially to an error page - this is normal)")
        print(
            "Step 3: Copy the response URL - click into the address bar and copy all."
        )
        response = input("Step 4: paste the response here: ").strip()

        # Verify the state which ensures the response is for this request
        return_state = re.search("&state=([^&]+)", response)
        if return_state:
            if return_state.group(1) != state:
                raise Exception(
                    "Response Error : Response 'state' did not correspond to request, typically occurs when reusing an old authorization url"
                )
        else:
            warnings.warn(
                "Response Warn : Response 'state' was not in returned url, response could not be confirmed as being for this request",
                stacklevel=2,
            )

        # Extract the code from the response
        authorization_code_re = re.search("code=([^&]+)", response)
        if authorization_code_re is None:
            raise Exception("The response did not contain an authorization code.")
        authorization_code = authorization_code_re.group(1)

        # Return the authorization code to be used to get tokens
        return authorization_code

    def _create_headers(self) -> None:
        """INTENRAL: Create headers for the http request to the Graph API."""
        self._headers = {
            "Accept": "*/*",
            "Authorization": "Bearer " + self._access_token,
        }

    @token_required
    def _get_drive_details(self) -> None:
        """INTERNAL: Gets the drive details"""
        # Generate request url
        request_url = self._API_URL + "me/drive/"
        response = requests.get(request_url, headers=self._headers)
        response_data = json.loads(response.text)
        # Set drive details
        self._drive_id = response_data.get("id")
        self._drive_name = response_data.get("name")
        self._drive_type = response_data.get("driveType")
        response_data_user = response_data.get("owner", {}).get("user", {})
        self._owner_id = response_data_user.get("id")
        self._owner_email = response_data_user.get("email")
        self._owner_name = response_data_user.get("displayName")
        response_data_quota = response_data.get("quota", {})
        self._quota_used = response_data_quota.get("used")
        self._quota_remaining = response_data_quota.get("remaining")
        self._quota_total = response_data_quota.get("total")

    @token_required
    def get_usage(
        self, unit: str = "gb", refresh: bool = False, verbose: bool = False
    ) -> Tuple[float, float, str]:
        """Get the current usage and capacity of the connected OneDrive.
        Keyword arguments:
            unit (str) -- unit to return value ["b", "kb", "mb", "gb"] (default = "gb")
            refresh (bool) -- refresh the usage data (default = False)
            verbose (bool) -- print the usage (default = False)
        Returns:
            used (float) -- storage used in unit requested
            capacity (float) -- storage capacity in unit requested
            units (str) -- unit of usage
        """
        if refresh:
            self._get_drive_details()
        # Read usage values
        used = self._quota_used
        capacity = self._quota_total
        # Convert to requested unit unit
        if unit == "gb":
            used = round(used / (1024 * 1024 * 1024), 2)
            capacity = round(capacity / (1024 * 1024 * 1024), 2)
        elif unit == "mb":
            used = round(used / (1024 * 1024), 2)
            capacity = round(capacity / (1024 * 1024), 2)
        elif unit == "kb":
            used = round(used / (1024), 2)
            capacity = round(capacity / (1024), 2)
        else:
            unit = "b"
        # Print usage
        if verbose:
            print(
                f"Using {used} {unit} ({round(used*100/capacity, 2)}%) of total {capacity} {unit}."
            )
        # Return usage and capacity in requested units
        return used, capacity, unit

    @token_required
    def list_directory(
        self, folder_id: Optional[str] = None, verbose: bool = False
    ) -> List[Dict[Any, Any]]:
        """List the files and folders within the input folder/root of the connected OneDrive.
        Keyword arguments:
            folder_id (str) -- the item id of the folder to look into, None being the root directory (default = None)
            verbose (bool) -- print the items along with their ids (default = False)
        Returns:
            items (dict) -- details of all the items within the requested directory
        """
        # Check if folder id was provided and create the request url
        if folder_id:
            request_url = self._API_URL + "me/drive/items/" + folder_id + "/children"
        else:
            request_url = self._API_URL + "me/drive/root/children"
        # Make the Graph API request
        response = requests.get(request_url, headers=self._headers)
        # Validate request response and parse
        if response.status_code != 200:
            print(response.text)
            raise Exception(f"API Error! : {response.status_code}")
        items = json.loads(response.text)
        items = items.get("value", {})
        # Print the items in the directory along with their item ids
        if verbose:
            for entries in range(len(items)):
                print(items[entries]["name"], "| item-id >", items[entries]["id"])
        # Return the items dictionary
        return items

    @token_required
    def detail_item(self, item_id: str, verbose: bool = False) -> Dict[str, Any]:
        """Retrieves the metadata for an item.
        Positional arguments:
            item_id (str) -- item id of the folder or file
        Keyword arguments:
            verbose (bool) -- print the main parts of the item metadata (default = False)
        Returns:
            item_details (dict) -- metadata of the requested item
        """
        # Create request url based on input item id
        request_url = self._API_URL + "me/drive/items/" + item_id
        # Make the Graph API request
        response = requests.get(request_url, headers=self._headers)
        # Validate request response and parse
        if response.status_code != 200:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: item could not be detailed."
            )
        response_data = json.loads(response.text)
        # Print the item details
        if verbose:
            print("id:", response_data.get("id"))
            print("name:", response_data.get("name"))
            if "folder" in response_data:
                print("type:", "folder")
            else:
                print("type:", "file")
            print(
                "created:",
                response_data.get("createdDateTime"),
                "by:",
                response_data.get("createdBy", {}).get("user", {}).get("displayName"),
            )
            print(
                "modified:",
                response_data.get("lastModifiedDateTime"),
                "by:",
                response_data.get("lastModifiedBy", {})
                .get("user", {})
                .get("displayName"),
            )
            print("size:", response_data.get("size"))
            print("web url:", response_data.get("webUrl"))
        # Return the item details
        return response_data

    @token_required
    def make_folder(
        self,
        folder_name: str,
        parent_folder_id: Optional[str] = None,
        check_existing: bool = True,
        if_exists: str = "rename",
    ) -> str:
        """Creates a new folder within the input folder/root of the connected OneDrive.
        Positional arguments:
            folder_name (str) -- the name of the new folder
        Keyword arguments:
            parent_folder_id (str) -- the item id of the parent folder, None being the root directory (default = None)
            check_existing (bool) -- checks parent and returns folder_id if a matching folder already exists (default = True)
            if_exists (str) -- if check_existing is set to False; action to take if the new folder already exists [fail, replace, rename] (default = "rename")
        Returns:
            folder_id (str) -- newly created folder item id
        """
        # Create request url based on input parent folder
        if parent_folder_id:
            request_url = (
                self._API_URL + "me/drive/items/" + parent_folder_id + "/children"
            )
        else:
            request_url = self._API_URL + "me/drive/root/children"
        # Set conflict behaviour
        if if_exists == "fail":
            conflictBehavior = "fail"
        elif if_exists == "replace":
            conflictBehavior = "replace"
        elif if_exists == "rename":
            conflictBehavior = "rename"
        else:
            raise Exception(
                "if_exists input value was not valid. Str type of values 'fail', 'replace', 'rename', are only accepted."
            )
        # Check if folder already exists
        if check_existing:
            items = self.list_directory(parent_folder_id)
            for i, entry in enumerate(items):
                if entry.get("name") == folder_name and "folder" in entry:
                    return entry["id"]

        # Create the request body
        body = {
            "name": folder_name,
            "folder": {},
            "@microsoft.graph.conflictBehavior": conflictBehavior,
        }
        # Make the Graph API request
        response = requests.post(request_url, headers=self._headers, json=body)
        # Validate request response and parse
        if response.status_code != 201:
            print(response.text)
            raise Exception(f"API Error {response.status_code}: folder was not created")
        response_data = json.loads(response.text)
        folder_id = response_data["id"]
        # Return the folder item id
        return folder_id

    @token_required
    def move_item(
        self, item_id: str, new_folder_id: str, new_name: Optional[str] = None
    ) -> Tuple[str, str]:
        """Moves an item (folder/file) within the connected OneDrive. Optionally rename an item at the same time.
        Positional arguments:
            item_id (str) -- item id of the folder or file to move
            new_folder_id (str) -- item id of the folder to shift the item to
        Keyword arguments:
            new_name (str) -- optional new item name with extension (default = None)
        Returns:
            item_id (str) -- item id of the folder or file that was moved, should match input item id
            folder_id (str) -- item id of the new parent folder, should match input folder id
        """
        # Create request url based on input item id that should be moved
        request_url = self._API_URL + "me/drive/items/" + item_id
        # Create the request body
        body: Dict[str, Any] = {"parentReference": {"id": new_folder_id}}
        if new_name:
            body["name"] = new_name
        # Make the Graph API request
        response = requests.patch(request_url, headers=self._headers, json=body)
        # Validate request response and parse
        if response.status_code != 200:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: item could not be moved."
            )
        response_data = json.loads(response.text)
        item_id = response_data["id"]
        parent_folder_id = response_data["parentReference"]["id"]
        # Return the item id and parent folder id
        return item_id, parent_folder_id

    @token_required
    def copy_item(
        self,
        item_id: str,
        new_folder_id: str,
        new_name: Optional[str] = None,
        confirm_complete: bool = False,
    ) -> str:
        """Copies an item (folder/file) within the connected OneDrive server-side.
        Positional arguments:
            item_id (str) -- item id of the folder or file to copy
            new_folder_id (str) -- item id of the folder to copy the item to
        Keyword arguments:
            new_name (str) -- optional new item name with extension (default = None)
            confirm_complete (bool) -- waits for the copy operation to finish before returning (default = False)
        Returns:
            item_id (str) -- item id of the new item
        """
        # Create request url based on input item id that should be moved
        request_url = self._API_URL + "me/drive/items/" + item_id + "copy"
        # Create the request body
        body: Dict[str, Any] = {
            "parentReference": {"driveId": self._drive_id, "id": new_folder_id}
        }
        if new_name:
            body["name"] = new_name
        # Make the Graph API request
        response = requests.post(request_url, headers=self._headers, json=body)
        # Validate request response and parse
        if response.status_code != 202:
            raise Exception(f"API Error: item could not be copied.")
        monitor_url = response.url
        if confirm_complete:
            wait_duration = 10
            previous_complete = 0
            while True:
                sleep(wait_duration)
                response = requests.get(monitor_url)
                response_data = json.loads(response.text)
                if response_data["status"] == "completed":
                    break
                percentage_complete = response_data["percentageComplete"]
                wait_duration = (
                    100.0 / (percentage_complete - previous_complete) * wait_duration
                    - wait_duration
                )
                if wait_duration > 30:
                    wait_duration = 30
                previous_complete = percentage_complete

        new_item_id = response_data["resourceId"]
        # Return the item id
        return new_item_id

    @token_required
    def rename_item(self, item_id: str, new_name: str) -> str:
        """Renames an item (folder/file) without moving it within the connected OneDrive.
        Positional arguments:
            item_id (str) -- item id of the folder or file to rename
            new_name (str) -- new item name with extension
        Returns:
            item_name (str) -- new name of the folder or file that was renamed
        """
        # Create request url based on input item id that should be renamed
        request_url = self._API_URL + "me/drive/items/" + item_id
        # Create the request body
        body = {"name": new_name}
        # Make the Graph API request
        response = requests.patch(request_url, headers=self._headers, json=body)
        # Validate request response and parse
        if response.status_code != 200:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: item could not be renamed."
            )
        response_data = json.loads(response.text)
        item_name = response_data["name"]
        # Return the item id and parent folder id
        return item_name

    @token_required
    def delete_item(self, item_id: str, pre_confirm: bool = False) -> bool:
        """Deletes an item (folder/file) within the connected OneDrive. Potentially restorable in the OneDrive web browser client.
        Positional arguments:
            item_id (str) -- item id of the folder or file to be deleted
        Keyword arguments:
            pre_confirm (bool) -- confirm that you want to delete the file and not show the warning (default = False)
        Returns:
            confirmation (bool) -- True if item was deleted successfully
        """
        # Get the user to confirm that they want to delete
        if not pre_confirm:
            confirm = (
                input("Deleted files may not be restorable. Type 'delete' to confirm: ")
                .strip()
                .lower()
            )
            if confirm != "delete":
                print("Aborted.")
                return False
        # Create request url based on input item id that should be deleted
        request_url = self._API_URL + "me/drive/items/" + item_id
        # Make the Graph API request
        response = requests.delete(request_url, headers=self._headers)
        # Validate request response
        if response.status_code != 204:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: item could not be deleted."
            )
        # Return confirmation of deletion
        return True

    @token_required
    def download_file(self, item_id: str) -> str:
        """Downloads the file to the current working directory. Note folders cannot be downloaded.
        Positional arguments:
            item_id (str) -- item id of the file to be deleted
        Returns:
            file_name (str) -- returns the name of the file including extension
        """
        # Get item details
        details = self.detail_item(item_id)
        # Check that it is not a folder
        if "folder" in details:
            raise Exception(
                "Item id provided is for a folder which this function does not permit."
            )
        # Create request url based on input item id to be downloaded
        request_url = self._API_URL + "me/drive/items/" + item_id + "/content"
        # Make the Graph API request
        response = requests.get(
            request_url, headers=self._headers, allow_redirects=False
        )
        # Validate request response and parse
        if response.status_code != 302:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: item could not be downloaded."
            )
        download_url = response.headers["Location"]
        # Download the file
        file_name = details["name"]
        # file_name = url.split("/")[-1]
        with requests.get(download_url, stream=True) as r:
            r.raw.read = functools.partial(r.raw.read, decode_content=True)
            with open(file_name, "wb") as f:
                shutil.copyfileobj(r.raw, f, length=16 * 1024 * 1024)
        # Return the data in a text format
        return file_name

    @token_required
    def upload_file(
        self,
        file_path: Union[str, Path],
        new_file_name: Optional[str] = None,
        parent_folder_id: Optional[str] = None,
        if_exists: str = "rename",
    ) -> str:
        """Uploads a file to a particular folder with a provided file name.
        Delegates the upload task to the upload_large_file function if required.
            Positional arguments:
                file_path (str|Path) -- path of the file on the drive
            Keyword arguments:
                new_file_name (str) -- new name of the file as it should appear on OneDrive, without extension (default = None)
                parent_folder_id (str) -- item id of the folder to put the file within, if None then root (default = None)
                if_exists (str) -- action to take if the new folder already exists [fail, replace, rename] (default = "rename")
            Returns:
                item_id (str) -- item id of the newly uploaded file
        """
        # Set conflict behaviour
        if if_exists == "fail":
            conflict_behavior = "fail"
        elif if_exists == "replace":
            conflict_behavior = "replace"
        elif if_exists == "rename":
            conflict_behavior = "rename"
        else:
            raise Exception(
                "if_exists input value was not valid. Str type of values 'fail', 'replace', 'rename', are only accepted."
            )
        # Ensure file_path is a Path type and remove escape slashes
        if os.name == "nt":  # Windows
            file_path = str(file_path).replace("/", "")
        else:  # Other systems including posix (Mac, Linux)
            file_path = str(file_path).replace("\\", "")
        file_path = Path(file_path)
        # Set file name
        if new_file_name:
            file_name = new_file_name
        else:
            file_name = file_path.name
        # Checks the file size and delegates the task to the upload_large_file function if required
        file_size = os.path.getsize(file_path)
        basic_file_limit = 4 * 1024 * 1024
        if file_size > basic_file_limit:
            print(f"Large file, uploading in chunks")
            response = self.upload_large_file(
                file_path, file_name, parent_folder_id, if_exists
            )
            return response
        # Create request url based on input values
        if parent_folder_id:
            request_url = self._API_URL + "me/drive/items/" + parent_folder_id + ":/"
        else:
            request_url = self._API_URL + "me/drive/root:/"
        request_url += (
            file_name
            + ":/content"
            + "?@microsoft.graph.conflictBehavior="
            + conflict_behavior
        )
        # Open the file into memory
        print("Loading file")
        content = open(file_path, "rb")
        # Make the Graph API request
        print("Uploading file")
        response = requests.put(request_url, headers=self._headers, data=content)
        # Close file
        content.close()
        # Validate request response and parse
        if response.status_code != 201 and response.status_code != 200:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}: could not upload file: {file_path}"
            )
        response_data = json.loads(response.text)
        item_id = response_data["id"]
        # Return the file item id
        return item_id

    @token_required
    def upload_large_file(
        self,
        file_path: Union[str, Path],
        new_file_name: Optional[str] = None,
        parent_folder_id: Optional[str] = None,
        if_exists: str = "rename",
    ) -> str:
        """Uploads a file in chunks to a particular folder with a provided file name.
        Positional arguments:
            file_path (str|Path) -- path of the file on the drive
        Keyword arguments:
            new_file_name (str) -- new name of the file as it should appear on OneDrive, without extension (default = None)
            parent_folder_id (str) -- item id of the folder to put the file within, if None then root (default = None)
            if_exists (str) -- action to take if the new folder already exists [fail, replace, rename] (default = "rename")
        Returns:
            item_id (str) -- item id of the newly uploaded file
        """
        # Set conflict behaviour
        if if_exists == "fail":
            conflict_behavior = "fail"
        elif if_exists == "replace":
            conflict_behavior = "replace"
        elif if_exists == "rename":
            conflict_behavior = "rename"
        else:
            raise Exception(
                f"if_exists={if_exists} not valid. Str type of values 'fail', 'replace', 'rename', are only accepted."
            )
        # Ensure file_path is a Path type
        file_path = Path(file_path)
        # Set file name
        if new_file_name:
            file_name = new_file_name
        else:
            file_name = file_path.name
        # Create request url for the upload session
        if parent_folder_id:
            request_url = self._API_URL + "me/drive/items/" + parent_folder_id + ":/"
        else:
            request_url = self._API_URL + "me/drive/root:/"
        request_url += file_name + ":/createUploadSession"
        # Create request body for the upload session
        body = {
            "item": {
                "@odata.type": "microsoft.graph.driveItemUploadableProperties",
                "@microsoft.graph.conflictBehavior": conflict_behavior,
            }
        }
        # Make the Graph API request for the upload session
        print(f"Requesting upload session using url")
        response = requests.post(request_url, headers=self._headers)
        # Validate upload session request response and parse
        if response.status_code != 200:
            print(response.text)
            raise Exception(
                f"API Error {response.status_code}, could not upload file: {file_path}"
            )
        upload_url = json.loads(response.text)
        upload_url = upload_url["uploadUrl"]
        # Determine the upload file size and chunks
        file_size = os.path.getsize(file_path)
        chunk_size = 320 * 1024 * 10  # Has to be multiple of 320 kb
        no_of_uploads = -(-file_size // chunk_size)
        content_range_start = 0
        if file_size < chunk_size:
            content_range_end = file_size
        else:
            content_range_end = chunk_size - 1
        # Open the file pointer
        print("Loading file")
        data = open(file_path, "rb")
        # Run in a try block to capture user cancellation request
        try:
            # Start the upload in a loop for as long as there is data left to upload
            n = 0
            while data.tell() < file_size:
                # Print the upload status
                n += 1
                if n == 1:
                    print(f"Uploading chunk {n}/{no_of_uploads}")
                else:
                    print(
                        f"Uploading chunk {n}/{no_of_uploads}  (~{int((n-1)/no_of_uploads*100)}% complete)"
                    )
                # Upload chunks
                if (file_size - data.tell()) > chunk_size:
                    # Typical chunk upload
                    headers = {
                        "Content-Range": "bytes "
                        + str(content_range_start)
                        + "-"
                        + str(content_range_end)
                        + "/"
                        + str(file_size)
                    }
                    content = data.read(chunk_size)
                    response = requests.put(upload_url, headers=headers, data=content)
                    # Validate request response
                    if response.status_code != 202:
                        data.close()
                        response2 = requests.delete(upload_url)
                        raise Exception(
                            f"API Error {response.status_code}: could not upload chuck {n} of {no_of_uploads}"
                        )
                    # Calculate next chunk range
                    content_range_start = data.tell()
                    content_range_end = data.tell() + chunk_size - 1
                else:
                    # Final chunk upload
                    content_range_end = file_size - 1
                    headers = {
                        "Content-Range": "bytes "
                        + str(content_range_start)
                        + "-"
                        + str(content_range_end)
                        + "/"
                        + str(file_size)
                    }
                    content = data.read(chunk_size)
                    response = requests.put(upload_url, headers=headers, data=content)
                    print("Upload complete")
        except KeyboardInterrupt:
            # Upload cancelled, send delete request
            data.close()
            response2 = requests.delete(upload_url)
            print("Upload cancelled by user.")
        # Close the file
        data.close()
        # Validate request response and parse
        if response.status_code != 201 and response.status_code != 200:
            print(response.text)
            response2 = requests.delete(upload_url)
            raise Exception(
                f"API Error {response.status_code}: could not upload file: {file_path}"
            )
        response_data = json.loads(response.text)
        item_id = response_data["id"]
        # Return the file item id
        return item_id



"""Decorators to wrap Graph-OneDrive package methods and functions.
"""
from datetime import datetime
from functools import wraps
from typing import no_type_check


@no_type_check
def token_required(func):
    """INTERNAL: Graph-OneDrive decorator to check for and refresh the access token when calling methods."""

    @wraps(func)
    def wrapper_token(*args, **kwargs):
        # Set self from args to get instance expires attribute
        onedrive_instance = args[0]
        expires = onedrive_instance._access_expires
        # Get the current timestamp for comparison
        now = datetime.timestamp(datetime.now())
        # Refresh access token if expires does not exist or has expired
        if expires <= now:
            onedrive_instance._get_token()
            onedrive_instance._create_headers()
        # Fucnction that is wrapped runs here
        wrapped_func = func(*args, **kwargs)
        # After run do nothing
        # Return the wrapped function
        return wrapped_func

    return wrapper_token



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="graph_onedrive",
    install_requires=["requests"],  # To-do: specify a min version based on testing
)
